@@ -51,9 +51,9 @@
 #include "version.h"
 #include "config.h"
 
-extern int show_module_info(const char *filename, const char *fmtstr, int do_parameters);
-extern int show_module_info_32(const char *filename, const char *fmtstr, int do_parameters);
-extern int show_module_info_64(const char *filename, const char *fmtstr, int do_parameters);
+extern int show_module_info(const char *filename, const char *fmtstr);
+extern int show_module_info_32(const char *filename, const char *fmtstr);
+extern int show_module_info_64(const char *filename, const char *fmtstr);
 
 #if defined(COMMON_3264) && defined(ONLY_32)
 #define SHOW_MODULE_INFO show_module_info_32		/* 32 bit version */
@@ -139,40 +139,25 @@ static const char *get_modinfo_value(struct obj_file *f, const char *key)
  * second and subsequent lines are always indented.
  */
 
-static int append_modinfo_tag(struct obj_file *f, char *tag, char *def,
-			      char **out_str, int n, int multi_line)
+static void append_modinfo_tag(struct obj_file *f, char *tag, char *def,
+			      int multi_line)
 {
 	const char *p, *p1;
-	int extra = 0, l;
+	int l;
 
 	p = get_modinfo_value(f, tag);
 	if (!p && !def)
-		return -1;
+		return;
 
 	if (!p) {
 		p = def;
 		multi_line = 0;
 	}
 
-	if (multi_line) {
-		p1 = p;
-		extra = 2;
-		while ((p1 = strchr(p1, '\n'))) {
-			extra += 2;
-			++p1;
-		}
-	}
-
-	if (strlen(p)+extra < n)
-		n = strlen(p)+extra;
-
-	if (!multi_line || n < 2) {
-		strncpy(*out_str, p, n);
-		*out_str += n;
-	}
+	if (!multi_line)
+		fputs (p, stdout);
 	else {
-		*(*out_str)++ = '"';
-		n -= 2;
+		putchar ('"');
 		p1 = p;
 		while (*p1) {
 			p1 = strchr(p, '\n');
@@ -181,116 +166,13 @@ static int append_modinfo_tag(struct obj_file *f, char *tag, char *def,
 			else
 				p1 = p + strlen(p);
 			l = p1 - p;
-			if (l > n)
-				l = n;
-			strncpy(*out_str, p, l);
-			n -= l;
-			*out_str += l;
-			if (*(p1-1) == '\n' && n >= 2) {
-				strncpy(*out_str, "  ", 2);
-				*out_str += 2;
-				n -= 2;
-			}
+			fwrite (p, 1, l, stdout);
+			if (*(p1-1) == '\n')
+				fputs ("  ", stdout);
 			p = p1;
 		}
-		*(*out_str)++ = '"';
+		putchar ('"');
 	}
-
-	return 0;
-}
-
-/* Format the "format" string based on the module contents. */
-static char *format_query_string(struct obj_file *f, const char *format)
-{
-	static char buffer[1024];
-	int c;
-	const char *in_str = format;
-	char *out_str = buffer;
-	char *last_char = buffer + sizeof(buffer) - 1;
-
-	while (*in_str && out_str < last_char) {
-		/* Just copy normal characters into the output. */
-		if (*in_str != '%')
-		  {
-		    if (*in_str == '\\')
-		      c = convert_escape(*++in_str);
-		    else
-		      c = *in_str;
-
-		    ++in_str;
-
-		    *out_str++ = (char) c;
-		    continue;
-		  }
-		/* Ensure that the replacement is there. */
-		if (*++in_str == '\0') {
-			error("missing replacement");
-			return NULL;
-		}
-		/* Process the replacement as required. */
-		switch (*in_str++) {
-
-		case '%':
-			*out_str++ = '%';
-			break;
-
-		case 'a':
-			(void) append_modinfo_tag(f, "author", "<none>",
-				&out_str, last_char - out_str, 1);
-			break;
-
-		case 'd':
-			(void) append_modinfo_tag(f, "description", "<none>",
-				&out_str, last_char - out_str, 1);
-			break;
-
-		case 'n':
-			(void) append_modinfo_tag(f, "filename", "<none>",
-				&out_str, last_char - out_str, 0);
-			break;
-
-		case 'l':
-			(void) append_modinfo_tag(f, "license", "<none>",
-				&out_str, last_char - out_str, 1);
-			break;
-
-		case '{':{
-				char tag[128], *end = strchr(in_str, '}');
-				int multi_line;
-
-				/* Make sure the %{...} is formatted correctly. */
-				if (!end) {
-					error("unterminated %%{...} construct");
-					return NULL;
-				}
-				if (end - in_str > sizeof(tag) - 1) {
-					error("%%{...} construct is too large");
-					return NULL;
-				}
-				/* Copy out the tag name. */
-				memset(tag, 0, sizeof(tag));
-				strncpy(tag, in_str, end - in_str);
-				tag[end - in_str] = '\0';
-				multi_line = strcmp(tag, "author") == 0 ||
-					     strcmp(tag, "description") == 0 ||
-					     strcmp(tag, "license") == 0;
-
-				/* Append the tag's value if it exists. */
-				append_modinfo_tag(f, tag, "<none>", &out_str, last_char - out_str, multi_line);
-
-				/* Advance past the end of the replacement. */
-				in_str = end + 1;
-				break;
-			}
-
-		default:
-			error("%c: unknown replacement", in_str[-1]);
-			return NULL;
-		}		/* switch */
-	}			/* while */
-
-	*out_str = '\0';
-	return &buffer[0];
 }
 
 static void show_parameter(struct obj_file *f, char *key, char *value,
@@ -316,7 +198,7 @@ static void show_parameter(struct obj_file *f, char *key, char *value,
 	if (max < min)
 		printf("warning: parameter %s has max < min!\n", key);
 
-	printf("parm:        %s ", key);
+	printf("%s ", key);
 
 	switch (*p) {
 	case 'c':
@@ -381,11 +263,144 @@ static void show_parameter(struct obj_file *f, char *key, char *value,
 		}
 		printf("\"");
 	}
-
-	printf("\n");
 }
 
-static void show_module_parameters(struct obj_file *f)
+#define MAX_TAG_LEN 128
+
+/* Check the "format" string. */
+static int format_check_string(const char *format)
+{
+	const char *in_str = format;
+
+	while (*in_str) {
+		/* Just skip normal characters. */
+		if (*in_str != '%')
+		  {
+		    if (*in_str == '\\')
+		      convert_escape(*++in_str);
+		    ++in_str;
+		    continue;
+		  }
+		/* Ensure that the replacement is there. */
+		if (*++in_str == '\0') {
+			error("missing replacement");
+			return -1;
+		}
+		/* Process the replacement as required. */
+		switch (*in_str++) {
+
+		case '%':
+		case 'a':
+		case 'd':
+		case 'n':
+		case 'l':
+		case 'p':
+			break;
+
+		case '{':{
+				char *end = strchr(in_str, '}');
+
+				/* Make sure the %{...} is formatted correctly. */
+				if (!end) {
+					error("unterminated %%{...} construct");
+					return -1;
+				}
+				if (end - in_str > MAX_TAG_LEN - 1) {
+					error("%%{...} construct is too large");
+					return -1;
+				}
+				/* Advance past the end of the replacement. */
+				in_str = end + 1;
+				break;
+			}
+
+		default:
+			error("%c: unknown replacement", in_str[-1]);
+			return -1;
+		}		/* switch */
+	}			/* while */
+	return 0;
+}
+
+/* Format one line of the "format" string based on the module contents. */
+static void format_query_line(struct obj_file *f, const char *format,
+			      char *key, char *value, const char *desc)
+{
+	int c;
+	const char *in_str = format;
+
+	while (*in_str) {
+		/* Just copy normal characters into the output. */
+		if (*in_str != '%')
+		  {
+		    if (*in_str == '\\')
+		      c = convert_escape(*++in_str);
+		    else
+		      c = *in_str;
+
+		    ++in_str;
+
+		    putchar (c);
+		    if (c == '\n')
+			    break;
+		    continue;
+		  }
+		++in_str;
+		/* Process the replacement as required. */
+		switch (*in_str++) {
+
+		case '%':
+			putchar ('%');
+			break;
+
+		case 'a':
+			append_modinfo_tag(f, "author", "<none>", 1);
+			break;
+
+		case 'd':
+			append_modinfo_tag(f, "description", "<none>", 1);
+			break;
+
+		case 'n':
+			append_modinfo_tag(f, "filename", "<none>", 0);
+			break;
+
+		case 'l':
+			append_modinfo_tag(f, "license", "<none>", 1);
+			break;
+
+		case 'p':
+			show_parameter(f, key, value, desc);
+			break;
+
+		case '{':{
+				char tag[MAX_TAG_LEN];
+				char *end = strchr(in_str, '}');
+				int multi_line;
+
+				/* Copy out the tag name. */
+				memset(tag, 0, sizeof(tag));
+				strncpy(tag, in_str, end - in_str);
+				tag[end - in_str] = '\0';
+				multi_line = strcmp(tag, "author") == 0 ||
+					     strcmp(tag, "description") == 0 ||
+					     strcmp(tag, "license") == 0;
+
+				if (strcmp (tag, "parm") != 0)
+				        /* Append the tag's value if it exists. */
+					append_modinfo_tag(f, tag, "<none>", multi_line);
+				else
+					show_parameter(f, key, value, desc);
+
+				/* Advance past the end of the replacement. */
+				in_str = end + 1;
+				break;
+			}
+		}		/* switch */
+	}			/* while */
+}
+
+static void show_module_parameters(struct obj_file *f, const char *format)
 {
 	struct obj_section *sec;
 	char *ptr, *value, *n, *endptr;
@@ -415,7 +430,7 @@ static void show_module_parameters(struct obj_file *f)
 				strncat(desckey, ptr + 5, namelen - 5);
 				desckey[namelen + 5] = '\0';
 
-				show_parameter(f, pname, value + 1, get_modinfo_value(f, desckey));
+				format_query_line(f, format, pname, value + 1, get_modinfo_value(f, desckey));
 
 				free(pname);
 				free(desckey);
@@ -429,11 +444,36 @@ static void show_module_parameters(struct obj_file *f)
 	}
 }
 
-int SHOW_MODULE_INFO(const char *filename, const char *fmtstr, int do_parameters)
+/* Format the "format" string based on the module contents. */
+static void format_query_string(struct obj_file *f, const char *format)
+{
+	while (*format != 0) {
+		const char *p;
+		int parm;
+
+		parm = 0;
+		for (p = format; *p != 0; p++) {
+			if (p > format && p[-1] == '%' && *p == 'p')
+				parm = 1;
+			if (p >= format + 6 && memcmp (p - 6, "%{parm}", 7) == 0)
+				parm = 1;
+			if (p > format && p[-1] == '\n')
+				break;
+			if (p > format + 1 && p[-2] == '\\' && p[-1] == 'n')
+				break;
+		}
+		if (parm == 0)
+			format_query_line(f, format, NULL, NULL, NULL);
+		else
+		  show_module_parameters(f, format);
+		format = p;
+	}
+}
+
+int SHOW_MODULE_INFO(const char *filename, const char *fmtstr)
 {
 	int fp;
 	struct obj_file *f;
-	char *p;
 
 	/* Locate the file to be loaded.  */
 	if (!strchr(filename, '/') && !strchr(filename, '.')) {
@@ -454,25 +494,18 @@ int SHOW_MODULE_INFO(const char *filename, const char *fmtstr, int do_parameters
 		return -1;
 	gzf_close(fp);
 
-	if (fmtstr) {
-		p = format_query_string(f, fmtstr);
-		if (p)
-			fputs(p, stdout);
-	}
-	if (do_parameters) {
-		show_module_parameters(f);
-	}
+	format_query_string(f, fmtstr);
 	return 0;
 }
 
 /* For common 3264 code, add an overall show_module_info, in the 64 bit version. */
 #if defined(COMMON_3264) && defined(ONLY_64)
-int show_module_info(const char *filename, const char *fmtstr, int do_parameters)
+int show_module_info(const char *filename, const char *fmtstr)
 {
 	if (arch64())
-		return show_module_info_64(filename, fmtstr, do_parameters);
+		return show_module_info_64(filename, fmtstr);
 	else
-		return show_module_info_32(filename, fmtstr, do_parameters);
+		return show_module_info_32(filename, fmtstr);
 }
 #endif  /* defined(COMMON_3264) && defined(ONLY_64) */
 
@@ -514,33 +547,28 @@ int main(int argc, char *argv[])
 		{"help", 0, 0, 'h'},
 		{0, 0, 0, 0}
 	};
-	int do_parameters = 1, opt;
-	char *fmtstr = "filename:    %n\ndescription: %d\nauthor:      %a\nlicense:     %l\n";
+	int opt;
+	char *fmtstr = "filename:    %n\ndescription: %d\nauthor:      %a\nlicense:     %l\nparm:        %p\n";
 
 	while ((opt = getopt_long(argc, argv, "adlnq:pf:Vh", long_opts, NULL)) != EOF)
 		switch (opt) {
 		case 'a':
 			fmtstr = "%a\n";
-			do_parameters = 0;
 			break;
 		case 'd':
 			fmtstr = "%d\n";
-			do_parameters = 0;
 			break;
 		case 'l':
 			fmtstr = "%l\n";
-			do_parameters = 0;
 			break;
 		case 'n':
 			fmtstr = "%n\n";
-			do_parameters = 0;
 			break;
 		case 'f':
-			fmtstr = xstrdup(optarg);
-			do_parameters = 0;
+			fmtstr = optarg;
 			break;
 		case 'p':
-			do_parameters = 1;
+			fmtstr = "%p\n";
 			break;
 		case 'V':
 			fputs("modinfo (Linux modutils) " MODUTILS_VERSION "\n", stdout);
@@ -557,8 +585,10 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 	error_file = "modinfo";
+	if (format_check_string (fmtstr) != 0)
+	  exit (1);
 	while (optind < argc)
-		show_module_info(argv[optind++], fmtstr, do_parameters);
+		show_module_info(argv[optind++], fmtstr);
 
 	return 0;
 }
